import { ec as EC } from "elliptic";
import { encodeFunctionData, keccak256, } from "viem";
import { ENTRYPOINT_ABI, ERC20_ABI, SESSION_KEY_MANAGER_ABI, SMART_ACCOUNT_V1_ABI, SMART_ACCOUNT_V1_FACTORY_ABI, UNISWAP_V3_FACTORY_ABI, UNISWAP_V3_SWAP_ROUTER_ABI, } from "./abi";
import { DEX_CHAINS, ENTRYPOINT_ADDRESS, SMART_ACCOUNT_V1_CHAINS, } from "./addresses";
export class SmartAccountV1Provider {
    constructor(publicClient, factory, sessionKeyManager, dex) {
        Object.defineProperty(this, "SESSION_MANAGER_DOMAIN_NAME", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: "LongDefi Session Key Manager"
        });
        Object.defineProperty(this, "ecdsa", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new EC("secp256k1")
        });
        Object.defineProperty(this, "publicClient", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "dex", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "factory", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "sessionKeyManager", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        const dexChain = DEX_CHAINS[publicClient.chain.id];
        const smartAccountChain = SMART_ACCOUNT_V1_CHAINS[publicClient.chain.id];
        this.publicClient = publicClient;
        this.dex = dex || dexChain;
        this.factory =
            factory || (smartAccountChain && smartAccountChain.smartAccountFactoryV1);
        this.sessionKeyManager =
            sessionKeyManager ||
                (smartAccountChain && smartAccountChain.sessionKeyManager);
    }
    setDex(dex) {
        this.dex = dex;
    }
    setFactory(factory) {
        this.factory = factory;
    }
    setSessionKeyManager(sessionKeyManager) {
        this.sessionKeyManager = sessionKeyManager;
    }
    async createSessionKeyRequest(owner, smartAccountSalt) {
        if (!this.sessionKeyManager) {
            throw new Error("Session key manager not supported on this chain");
        }
        const sessionKeyPair = this.ecdsa.genKeyPair();
        const publicKey = sessionKeyPair.getPublic("hex");
        const kc = keccak256(`0x${publicKey}`).slice(2);
        const sessionKeyAddress = `0x${kc.slice(-40)}`;
        const domain = {
            name: this.SESSION_MANAGER_DOMAIN_NAME,
            version: "1",
            chainId: this.publicClient.chain.id,
            verifyingContract: this.sessionKeyManager,
        };
        const types = {
            Permit: [
                {
                    name: "owner",
                    type: "address",
                },
                {
                    name: "salt",
                    type: "uint256",
                },
                {
                    name: "sessionKey",
                    type: "address",
                },
                {
                    name: "nonce",
                    type: "uint256",
                },
            ],
        };
        const isAuthorized = await this.publicClient.readContract({
            abi: SESSION_KEY_MANAGER_ABI,
            address: this.sessionKeyManager,
            functionName: "isAuthorized",
            args: [owner, sessionKeyAddress],
        });
        if (isAuthorized) {
            throw new Error("Session key already authorized");
        }
        const nonce = await this.publicClient.readContract({
            abi: SESSION_KEY_MANAGER_ABI,
            address: this.sessionKeyManager,
            functionName: "nonces",
            args: [owner],
        });
        const request = {
            domain,
            types,
            primaryType: "Permit",
            message: {
                owner,
                salt: smartAccountSalt,
                sessionKey: sessionKeyAddress,
                nonce,
            },
        };
        return {
            sessionKey: {
                privateKey: sessionKeyPair.getPrivate("hex"),
                address: sessionKeyAddress,
            },
            request,
        };
    }
    async revokeSessionKeyRequest(owner, smartAccountSalt, pub, enc) {
        if (!this.sessionKeyManager) {
            throw new Error("Session key manager not supported on this chain");
        }
        const key = this.ecdsa.keyFromPublic(pub, enc);
        const publicKey = key.getPublic("hex");
        const kc = keccak256(`0x${publicKey}`).slice(2);
        const sessionKey = `0x${kc.slice(-40)}`;
        const domain = {
            name: this.SESSION_MANAGER_DOMAIN_NAME,
            version: "1",
            chainId: this.publicClient.chain.id,
            verifyingContract: this.sessionKeyManager,
        };
        const types = {
            Revoke: [
                {
                    name: "owner",
                    type: "address",
                },
                {
                    name: "salt",
                    type: "uint256",
                },
                {
                    name: "sessionKey",
                    type: "address",
                },
                {
                    name: "nonce",
                    type: "uint256",
                },
            ],
        };
        const isAuthorized = await this.publicClient.readContract({
            abi: SESSION_KEY_MANAGER_ABI,
            address: this.sessionKeyManager,
            functionName: "isAuthorized",
            args: [owner, sessionKey],
        });
        if (isAuthorized) {
            throw new Error("Session key already authorized");
        }
        const nonce = await this.publicClient.readContract({
            abi: SESSION_KEY_MANAGER_ABI,
            address: this.sessionKeyManager,
            functionName: "nonces",
            args: [owner],
        });
        const request = {
            domain,
            types,
            primaryType: "Revoke",
            message: {
                owner,
                salt: smartAccountSalt,
                sessionKey,
                nonce,
            },
        };
        return request;
    }
    async createSwapRequest(createSwapRequestInput) {
        if (!createSwapRequestInput.smartAccount &&
            !createSwapRequestInput.initSmartAccountInput) {
            throw new Error("Must provide either `smartAccount` or `initSmartAccountInput`");
        }
        if (!this.dex) {
            throw new Error("Dex not supported");
        }
        if (!this.factory) {
            throw new Error("Factory not supported");
        }
        const dexName = createSwapRequestInput.dex;
        const dex = this.dex[dexName];
        if (!dex) {
            throw new Error(`${dexName} not supported`);
        }
        let smartAccount = createSwapRequestInput.smartAccount;
        let initCode = "0x";
        const smartAccountFactory = this.factory;
        const initSmartAccountInput = createSwapRequestInput.initSmartAccountInput;
        if (initSmartAccountInput) {
            const { owner, salt } = initSmartAccountInput;
            const address = await this.getSmartAccountAddress(owner, salt);
            const bytecode = await this.publicClient.getBytecode({ address });
            if (bytecode && bytecode !== "0x") {
                throw new Error("Smart account already exists");
            }
            const initCodeCallData = encodeFunctionData({
                abi: SMART_ACCOUNT_V1_FACTORY_ABI,
                functionName: "createAccount",
                args: [owner, salt],
            });
            initCode = (smartAccountFactory.toLowerCase() +
                initCodeCallData.slice(2));
            smartAccount = address.toLowerCase();
        }
        if (!smartAccount) {
            throw new Error("Smart account not found");
        }
        const { tokenIn, tokenOut, fee, recipient, deadline, amountIn, amountOutMinimum, sqrtPriceLimitX96, } = createSwapRequestInput.swapInput;
        const approveCallData = encodeFunctionData({
            abi: ERC20_ABI,
            functionName: "approve",
            args: [dex.swapRouter, amountIn],
        });
        const exactInputSingleCallData = encodeFunctionData({
            abi: UNISWAP_V3_SWAP_ROUTER_ABI,
            functionName: "exactInputSingle",
            args: [
                {
                    tokenIn,
                    tokenOut,
                    fee,
                    recipient: recipient || smartAccount,
                    deadline,
                    amountIn,
                    amountOutMinimum,
                    sqrtPriceLimitX96: sqrtPriceLimitX96 || BigInt(0),
                },
            ],
        });
        const dest = [tokenIn, dex.swapRouter];
        const value = [];
        const func = [approveCallData, exactInputSingleCallData];
        const executeBatchCallData = encodeFunctionData({
            abi: SMART_ACCOUNT_V1_ABI,
            functionName: "executeBatch",
            args: [dest, value, func],
        });
        const suffixId = createSwapRequestInput.orderId;
        const nonce = await this.getNonceForSmartAccount(smartAccount, dex.factory, tokenIn, tokenOut, fee, suffixId);
        const callGasLimit = await this.calculateExecuteBatchGasLimit(smartAccount, dest, value, func);
        const verificationGasLimit = await this.calculateVerificationGasLimitV1(initCode !== "0x");
        const preVerificationGas = BigInt(0);
        let maxFeePerGas = BigInt(0);
        let maxPriorityFeePerGas = BigInt(0);
        if (!createSwapRequestInput.gasless) {
            maxFeePerGas = BigInt(30e9);
            maxPriorityFeePerGas = BigInt(2e9);
        }
        const userOpWithoutSign = {
            sender: smartAccount,
            nonce,
            initCode,
            callData: executeBatchCallData,
            callGasLimit,
            verificationGasLimit,
            preVerificationGas,
            maxFeePerGas,
            maxPriorityFeePerGas,
            paymasterAndData: "0x",
            signature: "0x",
        };
        const userOpHash = await this.publicClient.readContract({
            abi: ENTRYPOINT_ABI,
            address: ENTRYPOINT_ADDRESS,
            functionName: "getUserOpHash",
            args: [userOpWithoutSign],
        });
        return { smartAccount, userOpHash, request: userOpWithoutSign };
    }
    async getSmartAccountAddress(owner, salt) {
        if (!this.factory) {
            throw new Error("Factory not supported");
        }
        const address = await this.publicClient.readContract({
            abi: SMART_ACCOUNT_V1_FACTORY_ABI,
            address: this.factory,
            functionName: "getAddress",
            args: [owner, salt],
        });
        return address;
    }
    async getNonceForSmartAccount(smartAccount, dexFactory, tokenIn, tokenOut, fee, suffixId) {
        const poolAddr = await this.publicClient.readContract({
            abi: UNISWAP_V3_FACTORY_ABI,
            address: dexFactory,
            functionName: "getPool",
            args: [tokenIn, tokenOut, fee],
        });
        const formattedPoolAddr = poolAddr.toLowerCase();
        const keyPrefix = BigInt(formattedPoolAddr) << BigInt(32);
        if (!(suffixId < 2 ** 31)) {
            throw new Error("Suffix id must be less than 2^31");
        }
        const zeroForOne = tokenIn.toLowerCase() < tokenOut.toLowerCase() ? 1n : 0n;
        const keySuffix = (zeroForOne << 31n) | BigInt(suffixId);
        const key = keyPrefix | keySuffix;
        const nonce = await this.publicClient.readContract({
            abi: ENTRYPOINT_ABI,
            address: ENTRYPOINT_ADDRESS,
            functionName: "getNonce",
            args: [smartAccount, key],
        });
        return nonce;
    }
    async calculateVerificationGasLimitV1(initSmartAccount) {
        if (initSmartAccount) {
            return BigInt(300000);
        }
        else {
            return BigInt(50000);
        }
    }
    async calculateExecuteBatchGasLimit(smartAccount, dest, value, func) {
        try {
            const estimatedGas = await this.publicClient.estimateContractGas({
                abi: SMART_ACCOUNT_V1_ABI,
                address: smartAccount,
                functionName: "executeBatch",
                args: [dest, value, func],
                account: ENTRYPOINT_ADDRESS,
            });
            const baseGas = BigInt(21000);
            if (estimatedGas < 30000) {
                return BigInt(500000);
            }
            return estimatedGas - baseGas;
        }
        catch (error) {
            return BigInt(200000);
        }
    }
}
//# sourceMappingURL=smartAccountV1Provider.js.map